/****************************************************************************\
**            XFDES.C - FAST DES ENCRYPTION ALOGRITHM FUNCTION              **
******************************************************************************
**             VAX/FDES Routines Modified By Doctor Dissector               **
******************************************************************************
** Copyright (c) 1991, By Doctor Dissector            Last Update: 06/01/91 **
\****************************************************************************/

/*
** XFDES - Version 5.00
** Copyright 1991 By Doctor Dissector
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 1, or (at your option)
** any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*=[ Copyright ]============================================================*/

char xfdes_c_msg[]="Source: xfdes.c, Copyright (c)1991 By Doctor Dissector";

/*=[ Include Files ]========================================================*/

#include "xfdes.h"

/*=[ Global Variables ]=====================================================*/

/*
** Key schedule. Alternating longs with low and high bits of key.
** Low and high 24 bits of keys are stored alternately.
*/
static sbpb24 KS[32];

/* Tables that combine the S and P operations. */
static sbpb24 S0L[64], S1L[64], S2L[64], S3L[64],
              S4L[64], S5L[64], S6L[64], S7L[64],
              S0H[64], S1H[64], S2H[64], S3H[64],
              S4H[64], S5H[64], S6H[64], S7H[64];

/*
** It is slightly faster to indirect through this table than to specify
** the desired table directly.
*/
static sbpb24 *stt[] =
 {
  S0L, S0H,
  S1L, S1H,
  S2L, S2H,
  S3L, S3H,
  S4L, S4H,
  S5L, S5H,
  S6L, S6H,
  S7L, S7H,
 };

#if defined(XFDES_KILLER_CRACKER)
/*
** For XFDES under Killer Cracker, the following code has been added to
** save the pre-initialized block and the pre-initialized KS values
*/
static obpb1 oldblock[66];
static sbpb24 KSB[32];
#endif

/*===========================================================================*/

/*
**    function: sbpb24 tfTOsixbit(ebpb24 tf)
** description: convert unsl in twenty-four bit contiguous format to six
**              bits per byte format
**      result: result of conversion
*/

sbpb24 tfTOsixbit(tf)
    ebpb24 tf;
{
    return((sbpb24) (((tf & 0x000003FL))        | ((tf & 0x0000FC0L) << 0x2) |
                     ((tf & 0x003F000L) << 0x4) | ((tf & 0x0FC0000L) << 0x6)));
}

/*===========================================================================*/

/*
**    function: ebpb24 sixbitTOtf(sbpb24 sb)
** description: convert unsl in six bits per byte format to twenty-four
**              bit contiguous format
**      return: result of conversion
*/

ebpb24 sixbitTOtf(sb)
    sbpb24 sb;
{
    return((ebpb24) (((sb & 0x0000003FL))       | ((sb >> 0x2) & 0x0000FC0L) |
                     ((sb >> 0x4) & 0x003F000L) | ((sb >> 0x6) & 0x0FC0000L)));
}

/*===========================================================================*/

/*
**    function: void fsetkey(obpb1 *key)
** description: set up the key schedule from the key
**      return: none
*/

void fsetkey(key)
    obpb1 *key;
{
    reg int   i, j, k;
        obpb1 C[28], D[28], s, t;

    /*
    ** Generate C and D by permuting the key.  The low order bit of each
    ** 8-bit char is not used, so C and D are only 28 bits apiece
    */
    for(i=27; i>(-1); i--) {
        C[i]=key[PC1_C[i]];
        D[i]=key[PC1_D[i]];
    }
    /*
    ** Generate Ki by rotating C and D according to schedule and pick up
    ** a permutation using PC2
    */
    for(i=0; i<32; i+=2) {
        for(k=0; k<shifts[i]; k++) {     /* rotate */
            s=C[0];
            t=D[0];
            for(j=0; j<27; j++) {
                C[j]=C[j+1];
                D[j]=D[j+1];
            }
            C[27]=s;
            D[27]=t;
        }
        /* Get Ki, note C and D are concatenated */
        KS[i]=0;
        KS[i+1]=0;
        for(j=0; j<24; j++) {
            KS[i] |= ((sbpb24) C[PC2_C[j]] << j);
            KS[i+1] |= ((sbpb24) D[PC2_D[j]] << j);
        }
        KS[i]=tfTOsixbit(KS[i]);
        KS[i+1]=tfTOsixbit(KS[i+1]);
    }
}

/*===========================================================================*/

/*
**    function: void xform(sbpb24 *quarters, sbpb24 saltvalue)
** description: encrypt a block
**      return: none
*/

void xform(quarters, saltvalue)
    sbpb24 *quarters, saltvalue;
{
    reg unsb   i;
    reg sbpb6  *dp;
    reg sbpb24 **stp, *kp, k;
        sbpb24 Dl, Dh, Rl=0, Rh=0, Ll=0, Lh=0, t;

    for(i=25; i>0; i--) {
        for(kp=KS; kp<(&KS[32]); ) {
            k=((Rl ^ Rh) & saltvalue);
            Dl=(k ^ (Rl ^ (*kp++)));
            Dh=(k ^ (Rh ^ (*kp++)));
            stp=((sbpb24 **) stt);
            dp=((sbpb6 *) &Dl);
            Ll ^= (*stp++)[(*dp)];
            Lh ^= (*stp++)[(*dp++)];
            Ll ^= (*stp++)[(*dp)];
            Lh ^= (*stp++)[(*dp++)];
            Ll ^= (*stp++)[(*dp)];
            Lh ^= (*stp++)[(*dp++)];
            Ll ^= (*stp++)[(*dp)];
            Lh ^= (*stp++)[(*dp)];
            dp=((sbpb6 *) &Dh);
            Ll ^= (*stp++)[(*dp)];
            Lh ^= (*stp++)[(*dp++)];
            Ll ^= (*stp++)[(*dp)];
            Lh ^= (*stp++)[(*dp++)];
            Ll ^= (*stp++)[(*dp)];
            Lh ^= (*stp++)[(*dp++)];
            Ll ^= (*stp++)[(*dp)];
            Lh ^= (*stp++)[(*dp)];
            k=((Ll ^ Lh) & saltvalue);
            Dl=(k ^ (Ll ^ (*kp++)));
            Dh=(k ^ (Lh ^ (*kp++)));
            stp=((sbpb24 **) stt);
            dp=((sbpb6 *) &Dl);
            Rl ^= (*stp++)[(*dp)];
            Rh ^= (*stp++)[(*dp++)];
            Rl ^= (*stp++)[(*dp)];
            Rh ^= (*stp++)[(*dp++)];
            Rl ^= (*stp++)[(*dp)];
            Rh ^= (*stp++)[(*dp++)];
            Rl ^= (*stp++)[(*dp)];
            Rh ^= (*stp++)[(*dp)];
            dp=((sbpb6 *) &Dh);
            Rl ^= (*stp++)[(*dp)];
            Rh ^= (*stp++)[(*dp++)];
            Rl ^= (*stp++)[(*dp)];
            Rh ^= (*stp++)[(*dp++)];
            Rl ^= (*stp++)[(*dp)];
            Rh ^= (*stp++)[(*dp++)];
            Rl ^= (*stp++)[(*dp)];
            Rh ^= (*stp++)[(*dp)];
        }
        t=Ll;
        Ll=Rl;
        Rl=t;
        t=Lh;
        Lh=Rh;
        Rh=t;
    }
    (*quarters++)=sixbitTOtf(Ll);
    (*quarters++)=sixbitTOtf(Lh);
    (*quarters++)=sixbitTOtf(Rl);
    (*quarters)=sixbitTOtf(Rh);
}

/*===========================================================================*/

/*
**    function: void toBA64F(reg sbpb24 *quarters, reg obpb1 *onebits64);
** description: creates table tmpE and executes final permutation
**      return: none
*/

void toBA64F(quarters, onebits64)
    sbpb24 *quarters;
    reg obpb1 *onebits64;
{
    reg    obpb1  *p, j;
    reg    sbpb24 quarter;
    static obpb1  tmpE[96];

    p=tmpE;
    quarter=(*quarters++);
    for(j=0; j<24; j++)
        (*p++)=((quarter >> j) & 0x1);
    quarter=(*quarters++);
    for(j=0; j<24; j++)
        (*p++)=((quarter >> j) & 0x1);
    quarter=(*quarters++);
    for(j=0; j<24; j++)
        (*p++)=((quarter >> j) & 0x1);
    quarter=(*quarters);
    for(j=0; j<24; j++)
        (*p++)=((quarter >> j) & 0x1);
    for(p=FP; (*p); )
        (*onebits64++)=tmpE[(*p++)];
}

/*===========================================================================*/

/*
**    function: char *fcrypt(unsb *pw, char *salt)
** description: main encryption engine
**      return: encrypted <pw>
*/

char *fcrypt(pw, salt)
    unsb *pw;
    char *salt;
{
    reg    unsb   *tpw;
    reg    obpb1  c, i, j;
    static char   iobuf[16];
    static sbpb24 out96[4];
           obpb1  block[66];

#if defined(XFDES_KILLER_CRACKER)
    if (!sameword) {
#endif

    tpw=pw;
    for(i=0; i<66; i++)
        block[i]=0;
    for(i=0; (*tpw) && (i<64); tpw++, i++) {
        for(j=0; j<7; j++, i++)
            block[i]=(((*tpw) >> (6-j)) & 0x1);
    }
    fsetkey(block);

#if defined(XFDES_KILLER_CRACKER)
    memcpy(oldblock, block, 66);
    memcpy(KSB, KS, 32);
    }
    else {
        memcpy(block, oldblock, 66);
        memcpy(KS, KSB, 32);
    }
#endif

    iobuf[0]=(*salt++);
    iobuf[1]=(*salt);
    xform(out96, tfTOsixbit((sbpb24) (SALT1[iobuf[0]] | SALT2[iobuf[1]])));
    toBA64F(out96, block);
    for(i=0; i<11; i++) {
        c=0;
        for(j=0; j<6; j++)
            c=((c << 0x1) | block[(6*i)+j]);
        iobuf[i+2]=CM[c];
    }
    iobuf[i+2]=0;
    if (!iobuf[1])
        iobuf[1]=iobuf[0];
    return(iobuf);
}

/*===========================================================================*/

/*
**    function: fbpb4 lookupS(unsl tableno, sbpb6R t6bits)
** description: look-up an s-box entry
**      return: value of s-box entry found
*/

fbpb4 lookupS(tableno, t6bits)
    unsl   tableno;
    sbpb6R t6bits;
{
    sbpb6  fixed6;

    fixed6=(((t6bits & 0x01) << 0x5) + ((t6bits & 0x02) << 0x2) +
            ((t6bits & 0x04))        + ((t6bits & 0x08) >> 0x2) +
            ((t6bits & 0x10) >> 0x4) + ((t6bits & 0x20) >> 0x1));
    return((fbpb4) SS[(int) tableno][(int) fixed6]);
}

/*===========================================================================*/

/*
**    function: void init(unsl tableno, sbpb24 *lowptr, sbpb24 *highptr)
** description: initializes des engine
**      return: none
*/

void init(tableno, lowptr, highptr)
    unsl tableno;
    sbpb24 *lowptr, *highptr;
{
    reg     fbpb4  i, k;
    reg     sbpb6R j;
    static  obpb1  tmp32[32], tmpP32[32], tmpE[48];

    for(j=0; j<64; j++) {
        k=lookupS(tableno, j);
        for(i=0; i<32; i++)
            tmp32[i]=0;
        for(i=0; i<4; i++)
            tmp32[(4*(int) tableno)+i] = ((obpb1) (k >> i) & 0x1);
        for(i=0; i<32; i++)
            tmpP32[i]=tmp32[P[i]];
        for(i=0; i<48; i++)
            tmpE[i]=tmpP32[(int) E[i]];
        lowptr[j]=0;
        highptr[j]=0;
        for(i=0; i<24; i++)
            lowptr[(int) j] |= ((unsl) tmpE[i] << i);
        for(k=0, i=24; i<48; i++, k++)
            highptr[(int) j] |= ((unsl) tmpE[i] << k);
        lowptr[j]=tfTOsixbit(lowptr[j]);
        highptr[j]=tfTOsixbit(highptr[j]);
    }
}

/*===========================================================================*/

/*
**    function: void init_des(void)
** description: initiates initialization of des engine, must be called
**              prior to fcrypt()
**      return: none
*/

void init_des()
{
    init((unsl) 0, S0L, S0H);
    init((unsl) 1, S1L, S1H);
    init((unsl) 2, S2L, S2H);
    init((unsl) 3, S3L, S3H);
    init((unsl) 4, S4L, S4H);
    init((unsl) 5, S5L, S5H);
    init((unsl) 6, S6L, S6H);
    init((unsl) 7, S7L, S7H);
}
