/****************************************************************************\
**            XFDES.H - FAST DES ENCRYPTION ALOGRITHM FUNCTION              **
******************************************************************************
**             VAX/FDES Routines Modified By Doctor Dissector               **
******************************************************************************
** Copyright (c) 1991, By Doctor Dissector            Last Update: 06/01/91 **
\****************************************************************************/

/*
** XFDES - Version 5.00
** Copyright 1991 By Doctor Dissector
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 1, or (at your option)
** any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*=[ Copyright ]============================================================*/

char xfdes_h_msg[]="Source: xfdes.h, Copyright (c)1991 By Doctor Dissector";

/*=[ Definitions ]==========================================================*/

#define reg     register
#define uns     unsigned
#define unsb    uns char
#define unsl    uns long

/*
** Types for the different ways to represent DES bit patterns
** Bits are *always* right justified within fields
*/

#define obpb1   unsb    /* One bit per byte */
#define sbpb6   unsb    /* Six bits per byte, 6 held */
#define sbpb6R  unsb    /* Six bits per byte, reversed order, 6 held */
#define sbpb24  unsl    /* Six bits per byte, 24 held */
#define ebpb24  unsl    /* Eight bits per byte, 24 held */
#define fbpb4   unsb    /* Four bits per byte, 4 held */
#define fbpb4R  unsb    /* Four bits per byte, reversed order, 4 held */

/*=[ Data Tables ]==========================================================*/

/* Final permutation */
static obpb1 FP[] =
 {
  58,10,70,22,82,34,94,46,57, 9,69,21,81,33,93,45,
  56, 8,68,20,80,32,92,44,55, 7,67,19,79,31,91,43,
  52, 4,64,16,76,28,88,40,51, 3,63,15,75,27,87,39,
  50, 2,62,14,74,26,86,38,49, 1,61,13,73,25,85,37,
   0,                   /* Zero marks the end of the table */
 };

/*
** Permuted-choice 1 from the key bits to yield C and D
** Note that bits 8,16... are left out, they are intended
** for a parity check, not zero based
*/
static unsb PC1_C[] =
 {
  56,48,40,32,24,16, 8,
   0,57,49,41,33,25,17,
   9, 1,58,50,42,34,26,
  18,10, 2,59,51,43,35,
 };

static unsb PC1_D[] =
 {
  62,54,46,38,30,22,14,
   6,61,53,45,37,29,21,
  13, 5,60,52,44,36,28,
  20,12, 4,27,19,11, 3,
 };

/*
** Permuted-choice 2, pick out bits from
** CD array to generate key schedule
*/
static unsb PC2_C[] =
 {
  13,16,10,23, 0, 4,
   2,27,14, 5,20, 9,
  22,18,11, 3,25, 7,
  15, 6,26,19,12, 1,
 };

static unsb PC2_D[] =
 {
  12,23, 2, 8,18,26,
   1,11,22,16, 4,19,
  15,20,10,27, 5,24,
  17,13,21, 7, 0, 3,
};

/* The E bit selection table */
static unsb E[] =
 {
  31, 0, 1, 2, 3, 4,
   3, 4, 5, 6, 7, 8,
   7, 8, 9,10,11,12,
  11,12,13,14,15,16,
  15,16,17,18,19,20,
  19,20,21,22,23,24,
  23,24,25,26,27,28,
  27,28,29,30,31, 0,
 };

/*
** P is a permutation on the selected
** combination of the current L and key
*/
static unsb P[] =
 {
  15, 6,19,20,
  28,11,27,16,
   0,14,22,25,
   4,17,30, 9,
   1, 7,23,13,
  31,26, 2, 8,
  18,12,29, 5,
  21,10, 3,24,
 };

/*
** The 8 selection functions mutated by the following function:
** (((r >> 3) & 0x1) + ((r >> 1) & 0x2) + ((r << 1) & 0x4) + ((r << 3) & 0x8))
*/
static fbpb4 SS[8][64] =
 {
   7, 2,11, 8, 4,15,13, 1,12, 5, 6, 3,10, 9, 0,14,
   0,15,14, 2, 7, 4,11, 8, 5, 6, 3,13, 9,10,12, 1,
   2, 8, 7, 1,11, 6, 4,13,15, 3, 9,14,12, 5,10, 0,
  15, 3, 1, 4, 2, 9, 8,14,10,13,12, 7, 5, 0, 6,11,

  15, 8, 1, 7, 6,13,12, 2, 9,14, 4,11, 3, 0,10, 5,
  12,11, 2,14,15, 4, 1, 7, 3, 0, 8, 5, 6, 9,13,10,
   0, 7,14,13, 5, 2,11, 8,10, 1, 3, 6, 9,12, 4,15,
  11, 1, 5, 8,12,15, 2, 4,13, 6,14, 3, 0,10, 7, 9,

   5, 0, 9, 7, 6,12,15,10, 8,11, 3,14,13, 2, 4, 1,
  11,14, 0, 9,12, 2, 6, 5, 4, 1,10, 7, 3,13,15, 8,
  11, 6, 2, 9, 1,15,12, 0,13, 8, 4, 3,10, 5, 7,14,
   8, 5,11, 0, 6, 9, 1,14, 2,15, 7,12,13,10, 4, 3,

  14,11, 7,12, 0, 6, 9, 5, 8, 4, 1,10,13, 3, 2,15,
  11, 1,13,10, 6,15, 0,12, 2,14, 4, 3, 8, 5, 7, 9,
   5, 6, 9, 0, 3,13,14,11,15, 8,12, 7,10, 4, 1, 2,
  12,15, 0, 6, 5, 8,11, 1, 9, 2,10,13, 3,14, 4, 7,

   4, 3, 2, 8,14, 5,13, 6, 1,10,12,15,11, 0, 7, 9,
   7,13, 4, 3, 2,14,11, 8,10, 0,15, 5,12, 9, 1, 6,
   2, 4, 8,13, 5,11,14, 1,15, 9, 3,10, 6,12, 0, 7,
  13, 1, 3,14, 8, 7, 4,11, 6,15, 0, 9, 5, 2,10,12,

   3, 8, 5,15, 9, 4, 6, 1, 0,11,12, 2, 7,14,10,13,
   5,15, 2, 4,14, 3, 9,10, 6, 8,11, 7, 0,13,12, 1,
   9, 7,15,10, 4, 1, 3,12,14, 0, 2, 5, 8,11,13, 6,
   2,12, 4, 3, 9,10,15, 5,13, 7, 8,14, 6, 0, 1,11,

   2,13, 4, 7,15, 0, 1,11,12, 3, 9,14,10, 5, 6, 8,
  11, 0,13,14, 2, 9, 8, 5, 7,12,10, 3, 4,15, 1, 6,
   8, 2,13,11, 3,12,14, 7, 5,15, 6, 1, 0,10, 9, 4,
   6,13,11, 1, 8, 2, 5,14, 9,10, 0,15, 7, 4,12, 3,

  11, 4, 1, 2, 6,15,13, 8, 5, 9,12, 7,10, 0, 3,14,
   8,15,11, 1, 5,12,14, 2, 3,10, 6,13, 0, 7, 9, 4,
  14,13, 2, 8, 9, 3, 7, 4, 0, 6, 5,11,15,12,10, 1,
   4, 8, 7,14, 2, 5, 1,11,15, 3, 9, 0,12,10, 6,13,
 };

/* Sequence of shifts used for the key schedule */
static unsb shifts[] =
 {
  1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
  1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1,
 };

/*
** Pre-calculated salt tables for the first and second characters
** of the salt string
*/
static sbpb24 SALT1[] =
 {
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   1,
     2,   3,   4,   5,   6,   7,   8,   9,
    10,  11,   5,   6,   7,   8,   9,  10,
    11,  12,  13,  14,  15,  16,  17,  18,
    19,  20,  21,  22,  23,  24,  25,  26,
    27,  28,  29,  30,  31,  32,  33,  34,
    35,  36,  37,  32,  33,  34,  35,  36,
    37,  38,  39,  40,  41,  42,  43,  44,
    45,  46,  47,  48,  49,  50,  51,  52,
    53,  54,  55,  56,  57,  58,  59,  60,
    61,  62,  63,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
 };

static sbpb24 SALT2[] =
 {
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,  64,
   128, 192, 256, 320, 384, 448, 512, 576,
   640, 704, 320, 384, 448, 512, 576, 640,
   704, 768, 832, 896, 960,1024,1088,1152,
  1216,1280,1344,1408,1472,1536,1600,1664,
  1728,1792,1856,1920,1984,2048,2112,2176,
  2240,2304,2368,2048,2112,2176,2240,2304,
  2368,2432,2496,2560,2624,2688,2752,2816,
  2880,2944,3008,3072,3136,3200,3264,3328,
  3392,3456,3520,3584,3648,3712,3776,3840,
  3904,3968,4032,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
     0,   0,   0,   0,   0,   0,   0,   0,
 };

/*
** Character mutation table during final extraction
** of encrypted data, as used by fcrypt
*/
static obpb1 CM[] =
 {
   46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 65, 66, 67, 68,
   69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84,
   85, 86, 87, 88, 89, 90, 97, 98, 99,100,101,102,103,104,105,106,
  107,108,109,110,111,112,113,114,115,116,117,118,119,120,121,122,
  123,124,125,126,127,128,129,130,131,132,133,134,135,136,137,138,
  139,140,141,142,143,144,145,146,147,148,149,150,151,152,153,154,
  155,156,157,158,159,160,161,162,163,164,165,166,167,168,169,170,
  171,172,173,174,175,176,177,178,179,180,181,182,183,184,185,186,
  187,188,189,190,191,192,193,194,195,196,197,198,199,200,201,202,
  203,204,205,206,207,208,209,210,211,212,213,214,215,216,217,218,
  219,220,221,222,223,224,225,226,227,228,229,230,231,232,233,234,
  235,236,237,238,239,240,241,242,243,244,245,246,247,248,249,250,
  251,252,253,254,255,  0,  1,  2,  3,  4,  5,  0,  1,  2,  3,  4,
    5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20,
   21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36,
   37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52,
 };
